@@ -8,7 +8,6 @@ import os
 import re
 import shutil
 import subprocess
-import urllib.parse
 import urllib.request
 import stat
 import sys
@@ -20,6 +19,8 @@ import colorama
 import jsonschema
 import yaml
 
+from . import util
+
 verbosity = False
 
 global_yaml_loader = yaml.SafeLoader
@@ -560,6 +561,22 @@ class Source(RequirementsMixin):
 	def subject_type(self):
 		return 'source'
 
+	@property
+	def has_variable_checkout_commit(self):
+		if 'git' in self._this_yml:
+			return 'branch' in self._this_yml and 'commit' not in self._this_yml
+		return False
+
+	def determine_variable_checkout_commit(self):
+		if 'git' in self._this_yml and 'branch' in self._this_yml and 'commit' not in self._this_yml:
+			out = subprocess.check_output(['git', 'show-ref', '-s', '--verify',
+					'refs/remotes/origin/' + self._this_yml['branch']],
+					cwd=self.source_dir, stderr=subprocess.DEVNULL).decode().splitlines()
+			assert len(out) == 1
+			return out[0]
+		else:
+			raise RuntimeError("Source {} does not have a variable checkout commit".format(self.name))
+
 	@property
 	def is_rolling_version(self):
 		return self._this_yml.get('rolling_version', False)
@@ -1469,7 +1486,13 @@ def checkout_src(cfg, src, settings):
 		init = subprocess.call(['git', 'show-ref', '--verify', '-q', 'HEAD'],
 				cwd=src.source_dir) != 0
 
+		commit_yml = cfg._commit_yml.get('commits', dict()).get(src.name, dict())
+		fixed_commit = commit_yml.get('fixed_commit', None)
+
 		if 'tag' in source:
+			if fixed_commit is not None:
+				raise RuntimeError("Commit of source {} cannot be fixed in bootstrap-commits.yml: source builds form a branch".format(src.name));
+
 			if not init and settings.reset == ResetMode.HARD_RESET:
 				subprocess.check_call(['git', 'reset', '--hard'], cwd=src.source_dir)
 			if init or settings.reset != ResetMode.NONE:
@@ -1482,6 +1505,11 @@ def checkout_src(cfg, src, settings):
 			commit = 'origin/' + source['branch']
 			if 'commit' in source:
 				commit = source['commit']
+				if fixed_commit is not None:
+					raise RuntimeError("Commit of source {} cannot be fixed in bootstrap-commits.yml: commit is already fixed in bootstrap.yml".format(src.name));
+			else:
+				if fixed_commit is not None:
+					commit = fixed_commit
 			if init or settings.reset != ResetMode.NONE:
 				subprocess.check_call(['git', 'checkout', '--no-track',
 						'-B', source['branch'], commit], cwd=src.source_dir)
@@ -1654,16 +1682,36 @@ def compile_tool_stage(cfg, stage):
 	stage.mark_as_compiled()
 
 def install_tool_stage(cfg, stage):
-	pkg = stage.pkg
-	src = cfg.get_source(pkg.source)
+	tool = stage.pkg
+	src = cfg.get_source(tool.source)
+
+	# Sanity checking: make sure that the rolling ID matches the expected one.
+	src = cfg.get_source(tool.source)
+	if src.is_rolling_version:
+		actual_rolling_id = src.determine_rolling_id()
+		try:
+			if src.rolling_id != actual_rolling_id:
+				raise RuntimeError("Rolling ID of tool {} does not match true rolling ID".format(tool.name))
+		except RollingIdUnavailableException:
+			pass
+	else:
+		actual_rolling_id = None
+
+	version = tool.compute_version(override_rolling_id=actual_rolling_id)
 
 	try_mkdir(cfg.tool_out_dir)
-#	try_rmtree(pkg.prefix_dir)
-	try_mkdir(pkg.prefix_dir)
+#	try_rmtree(tool.prefix_dir)
+	try_mkdir(tool.prefix_dir)
+
+	try_mkdir(os.path.join(tool.prefix_dir, 'xbstrap'))
+	with open(os.path.join(tool.prefix_dir, 'xbstrap/tool-metadata.yml'), 'w') as f:
+		f.write(yaml.safe_dump({
+			'version': version
+		}))
 
 	for step in stage.install_steps:
 		tool_pkgs = []
-		for dep_name in pkg.tool_dependencies:
+		for dep_name in tool.tool_dependencies:
 			tool_pkgs.append(cfg.get_tool_pkg(dep_name))
 
 		def substitute(varname):
@@ -1676,13 +1724,13 @@ def install_tool_stage(cfg, stage):
 			elif varname == 'THIS_SOURCE_DIR':
 				return src.source_dir
 			elif varname == 'THIS_BUILD_DIR':
-				return pkg.build_dir
+				return tool.build_dir
 			elif varname == 'PREFIX':
-				return pkg.prefix_dir
+				return tool.prefix_dir
 			elif varname.startswith('OPTION:'):
 				return cfg.get_option_value(varname[7:])
 
-		run_step(cfg, step, pkg.build_dir, substitute, tool_pkgs, pkg.virtual_tools)
+		run_step(cfg, step, tool.build_dir, substitute, tool_pkgs, tool.virtual_tools)
 
 	stage.mark_as_installed()
 
@@ -1912,6 +1960,45 @@ def archive_pkg(cfg, pkg):
 		for ent in os.listdir(pkg.staging_dir):
 			tar.add(os.path.join(pkg.staging_dir, ent), arcname=ent)
 
+def pull_pkg_pack(cfg, pkg):
+	from . import xbps_utils as _xbps_utils
+	repo_url = cfg._root_yml['repositories']['xbps']
+
+	# Download the repodata file.
+	rd_path = os.path.join(cfg.xbps_repository_dir, 'remote-x86_64-repodata')
+	rd_url = urllib.parse.urljoin(repo_url + '/', 'x86_64-repodata')
+	print('{}xbstrap{}: Downloading x86_64-repodata from {}'.format(
+			colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+			repo_url))
+	util.interactive_download(rd_url, rd_path)
+
+	# Find the package within the repodata's index file.
+	index = _xbps_utils.read_repodata(rd_path)
+	if pkg.name not in index:
+		raise RuntimeError("Package {} not found in remote repository".format(pkg.name))
+	assert 'pkgver' in index[pkg.name]
+
+	# Download the xbps file.
+	xbps_file = '{}.x86_64.xbps'.format(index[pkg.name]['pkgver'])
+	pkg_url = urllib.parse.urljoin(repo_url + '/', xbps_file)
+	print('{}xbstrap{}: Downloading {} from {}'.format(
+			colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+			xbps_file, repo_url))
+	util.interactive_download(pkg_url, os.path.join(cfg.xbps_repository_dir, xbps_file))
+
+	# Run xbps-rindex.
+	output = subprocess.DEVNULL
+	if verbosity:
+		output = None
+
+	args = ['xbps-rindex', '-fa',
+			os.path.join(cfg.xbps_repository_dir, xbps_file)
+	]
+	print("{}xbstrap{}: Running {}".format(
+			colorama.Style.BRIGHT, colorama.Style.RESET_ALL,
+			args))
+	subprocess.call(args, stdout=output)
+
 def run_task(cfg, task):
 	tools_required = []
 	for dep_name in task.tool_dependencies:
@@ -2016,11 +2103,12 @@ class Action(Enum):
 	INSTALL_PKG = 13
 	ARCHIVE_TOOL = 14
 	ARCHIVE_PKG = 15
-	RUN = 16
-	RUN_PKG = 17
-	RUN_TOOL = 18
-	WANT_TOOL = 19
-	WANT_PKG = 20
+	PULL_PKG_PACK = 16
+	RUN = 17
+	RUN_PKG = 18
+	RUN_TOOL = 19
+	WANT_TOOL = 20
+	WANT_PKG = 21
 
 Action.strings = {
 	Action.FETCH_SRC: 'fetch',
@@ -2038,6 +2126,7 @@ Action.strings = {
 	Action.INSTALL_PKG: 'install',
 	Action.ARCHIVE_TOOL: 'archive-tool',
 	Action.ARCHIVE_PKG: 'archive',
+	Action.PULL_PKG_PACK: 'pull-pack',
 	Action.RUN: 'run',
 	Action.RUN_PKG: 'run',
 	Action.RUN_TOOL: 'run',
@@ -2112,6 +2201,7 @@ class PlanItem:
 			Action.INSTALL_PKG: lambda s, c: s.check_if_installed(c),
 			Action.ARCHIVE_TOOL: lambda s, c: s.check_if_archived(c),
 			Action.ARCHIVE_PKG: lambda s, c: ItemState(missing=True),
+			Action.PULL_PKG_PACK: lambda s, c: ItemState(missing=True),
 			Action.RUN: lambda s, c: ItemState(missing=True),
 			Action.RUN_PKG: lambda s, c: ItemState(missing=True),
 			Action.RUN_TOOL: lambda s, c: ItemState(missing=True),
@@ -2281,6 +2371,9 @@ class Plan:
 		elif action == Action.ARCHIVE_PKG:
 			item.build_edges.add((action.BUILD_PKG, subject))
 
+		elif action == Action.PULL_PKG_PACK:
+			pass
+
 		elif action == Action.RUN:
 			add_implicit_pkgs()
 			add_pkg_dependencies(subject)
@@ -2594,6 +2687,8 @@ class Plan:
 					archive_tool(self._cfg, subject)
 				elif action == Action.ARCHIVE_PKG:
 					archive_pkg(self._cfg, subject)
+				elif action == Action.PULL_PKG_PACK:
+					pull_pkg_pack(self._cfg, subject)
 				elif action == Action.RUN:
 					run_task(self._cfg, subject)
 				elif action == Action.RUN_PKG:
